import math
import warnings
from typing import Union

import numpy as np

# ------------------ Rust Downsamplers ------------------
from tsdownsample._rust import _tsdownsample_rs  # type: ignore[attr-defined]

from .downsampling_interface import AbstractDownsampler, AbstractRustDownsampler


class MinMaxDownsampler(AbstractRustDownsampler):
    def __init__(self) -> None:
        super().__init__(_tsdownsample_rs.minmax)

    @staticmethod
    def _check_valid_n_out(n_out: int):
        AbstractRustDownsampler._check_valid_n_out(n_out)
        if n_out % 2 != 0:
            raise ValueError("n_out must be even")


class M4Downsampler(AbstractRustDownsampler):
    def __init__(self):
        super().__init__(_tsdownsample_rs.m4)

    @staticmethod
    def _check_valid_n_out(n_out: int):
        AbstractRustDownsampler._check_valid_n_out(n_out)
        if n_out % 4 != 0:
            raise ValueError("n_out must be a multiple of 4")


class LTTBDownsampler(AbstractRustDownsampler):
    def __init__(self):
        super().__init__(_tsdownsample_rs.lttb)


class MinMaxLTTBDownsampler(AbstractRustDownsampler):
    def __init__(self):
        super().__init__(_tsdownsample_rs.minmaxlttb)

    def downsample(
        self, *args, n_out: int, minmax_ratio: int = 30, parallel: bool = False, **_
    ):
        assert minmax_ratio > 0, "minmax_ratio must be greater than 0"
        return super().downsample(
            *args, n_out=n_out, parallel=parallel, ratio=minmax_ratio
        )


# ------------------ EveryNth Downsampler ------------------


class EveryNthDownsampler(AbstractDownsampler):
    def _downsample(
        self, x: Union[np.ndarray, None], y: np.ndarray, n_out: int, **_
    ) -> np.ndarray:
        if x is not None:
            name = self.__class__.__name__
            warnings.warn(
                f"x is passed to downsample method of {name}, but is not taken "
                "into account by the current implementation of the EveryNth algorithm."
            )
        step = max(1, math.ceil(len(y) / n_out))
        return np.arange(0, len(y), step)



# Store some global configuration for tests

import numpy as np

_core_supported_dtypes = [
    np.float32,
    np.float64,
    np.int16,
    np.int32,
    np.int64,
    np.uint16,
    np.uint32,
    np.uint64,
    np.datetime64,
    np.timedelta64,
]

supported_dtypes_x = _core_supported_dtypes
supported_dtypes_y = _core_supported_dtypes + [np.float16, np.int8, np.uint8, np.bool_]

_core_rust_primitive_types = ["f32", "f64", "i16", "i32", "i64", "u16", "u32", "u64"]

rust_primitive_types_x = _core_rust_primitive_types
rust_primitive_types_y = _core_rust_primitive_types + ["f16", "i8", "u8"]



"""AbstractDownsampler interface-class, subclassed by concrete downsamplers."""

__author__ = "Jeroen Van Der Donckt"

import re
import warnings
from abc import ABC, abstractmethod
from types import ModuleType
from typing import Callable, List, Optional, Tuple, Union

import numpy as np


class AbstractDownsampler(ABC):
    """AbstractDownsampler interface-class, subclassed by concrete downsamplers."""

    def __init__(
        self,
        x_dtype_regex_list: Optional[List[str]] = None,
        y_dtype_regex_list: Optional[List[str]] = None,
    ):
        self.x_dtype_regex_list = x_dtype_regex_list
        self.y_dtype_regex_list = y_dtype_regex_list

    def _supports_dtype(self, arr: np.ndarray, y: bool = True):
        dtype_regex_list = self.y_dtype_regex_list if y else self.x_dtype_regex_list
        # base case
        if dtype_regex_list is None:
            return

        for dtype_regex_str in dtype_regex_list:
            m = re.compile(dtype_regex_str).match(str(arr.dtype))
            if m is not None:  # a match is found
                return
        raise ValueError(
            f"{arr.dtype} doesn't match with any regex in {dtype_regex_list} "
            f"for the {'y' if y else 'x'}-data"
        )

    @staticmethod
    def _check_valid_downsample_args(
        *args,
    ) -> Tuple[Union[np.ndarray, None], np.ndarray]:
        if len(args) == 2:
            x, y = args
        elif len(args) == 1:
            x, y = None, args[0]
        else:
            raise ValueError(
                "downsample() takes 1 or 2 positional arguments but "
                f"{len(args)} were given"
            )
        # y must be 1D array
        if y.ndim != 1:
            raise ValueError("y must be 1D array")
        # x must be 1D array with same length as y or None
        if x is not None:
            if x.ndim != 1:
                raise ValueError("x must be 1D array")
            if len(x) != len(y):
                raise ValueError("x and y must have the same length")
        return x, y

    @staticmethod
    def _check_valid_n_out(n_out: int):
        if n_out <= 0:
            raise ValueError("n_out must be greater than 0")

    @abstractmethod
    def _downsample(
        self, x: Union[np.ndarray, None], y: np.ndarray, n_out: int, **kwargs
    ) -> np.ndarray:
        """Downsample the data in x and y.

        Returns
        -------
        np.ndarray
            The selected indices.
        """
        raise NotImplementedError

    def downsample(self, *args, n_out: int, **kwargs):  # x and y are optional
        """Downsample y (and x).

        Call signatures::
            downsample([x], y, n_out, **kwargs)


        Parameters
        ----------
        x, y : array-like
            The horizontal / vertical coordinates of the data points.
            *x* values are optional.
            These parameters should be 1D arrays.
            These arguments cannot be passed as keywords.
        n_out : int
            The number of points to keep.
        **kwargs
            Additional keyword arguments are passed to the downsampler.

        Returns
        -------
        np.ndarray
            The selected indices.
        """
        self._check_valid_n_out(n_out)
        x, y = self._check_valid_downsample_args(*args)
        self._supports_dtype(y, y=True)
        if x is not None:
            self._supports_dtype(x, y=False)
        return self._downsample(x, y, n_out, **kwargs)


# ------------------- Rust Downsample Interface -------------------
DOWNSAMPLE_F = "downsample"


# the following dtypes are supported by the rust downsamplers (x and y)
_rust_dtypes = [
    "float32",
    "float64",
    "uint16",
    "uint32",
    "uint64",
    "int16",
    "int32",
    "int64",
    "datetime64",
    "timedelta64",
]
# <= 8-bit x-dtypes are not supported as the range of the values is too small to require
# downsampling
_y_rust_dtypes = _rust_dtypes + ["float16", "int8", "uint8", "bool"]


class AbstractRustDownsampler(AbstractDownsampler, ABC):
    """RustDownsampler interface-class, subclassed by concrete downsamplers."""

    def __init__(self, resampling_mod: ModuleType):
        super().__init__(_rust_dtypes, _y_rust_dtypes)  # same for x and y
        self.rust_mod = resampling_mod

        # Store the single core sub module
        self.mod_single_core = self.rust_mod.scalar
        if hasattr(self.rust_mod, "simd"):
            # use SIMD implementation if available
            self.mod_single_core = self.rust_mod.simd

        # Store the multi-core sub module (if present)
        self.mod_multi_core = None  # no multi-core implementation (default)
        if hasattr(self.rust_mod, "simd_parallel"):
            # use SIMD implementation if available
            self.mod_multi_core = self.rust_mod.simd_parallel
        elif hasattr(self.rust_mod, "scalar_parallel"):
            # use scalar implementation if available (when no SIMD available)
            self.mod_multi_core = self.rust_mod.scalar_parallel

    @staticmethod
    def _switch_mod_with_y(
        y_dtype: np.dtype, mod: ModuleType, downsample_func: str = DOWNSAMPLE_F
    ) -> Callable:
        """The x-data is not considered in the downsampling

        Assumes equal binning.

        Parameters
        ----------
        y_dtype : np.dtype
            The dtype of the y-data
        mod : ModuleType
            The module to select the appropriate function from
        downsample_func : str, optional
            The name of the function to use, by default DOWNSAMPLE_FUNC.
        """
        # FLOATS
        if np.issubdtype(y_dtype, np.floating):
            if y_dtype == np.float16:
                return getattr(mod, downsample_func + "_f16")
            elif y_dtype == np.float32:
                return getattr(mod, downsample_func + "_f32")
            elif y_dtype == np.float64:
                return getattr(mod, downsample_func + "_f64")
        # UINTS
        elif np.issubdtype(y_dtype, np.unsignedinteger):
            if y_dtype == np.uint8:
                return getattr(mod, downsample_func + "_u8")
            elif y_dtype == np.uint16:
                return getattr(mod, downsample_func + "_u16")
            elif y_dtype == np.uint32:
                return getattr(mod, downsample_func + "_u32")
            elif y_dtype == np.uint64:
                return getattr(mod, downsample_func + "_u64")
        # INTS (need to be last because uint is subdtype of int)
        elif np.issubdtype(y_dtype, np.integer):
            if y_dtype == np.int8:
                return getattr(mod, downsample_func + "_i8")
            elif y_dtype == np.int16:
                return getattr(mod, downsample_func + "_i16")
            elif y_dtype == np.int32:
                return getattr(mod, downsample_func + "_i32")
            elif y_dtype == np.int64:
                return getattr(mod, downsample_func + "_i64")
        # DATETIME -> i64 (datetime64 is viewed as int64)
        # TIMEDELTA -> i64 (timedelta64 is viewed as int64)
        # BOOLS -> int8 (bool is viewed as int8)
        raise ValueError(f"Unsupported data type (for y): {y_dtype}")

    @staticmethod
    def _switch_mod_with_x_and_y(
        x_dtype: np.dtype, y_dtype: np.dtype, mod: ModuleType
    ) -> Callable:
        """The x-data is considered in the downsampling

        Assumes equal binning.

        Parameters
        ----------
        x_dtype : np.dtype
            The dtype of the x-data
        y_dtype : np.dtype
            The dtype of the y-data
        mod : ModuleType
            The module to select the appropriate function from
        """
        # FLOATS
        if np.issubdtype(x_dtype, np.floating):
            if x_dtype == np.float16:
                return AbstractRustDownsampler._switch_mod_with_y(
                    y_dtype, mod, f"{DOWNSAMPLE_F}_f16"
                )
            elif x_dtype == np.float32:
                return AbstractRustDownsampler._switch_mod_with_y(
                    y_dtype, mod, f"{DOWNSAMPLE_F}_f32"
                )
            elif x_dtype == np.float64:
                return AbstractRustDownsampler._switch_mod_with_y(
                    y_dtype, mod, f"{DOWNSAMPLE_F}_f64"
                )
        # UINTS
        elif np.issubdtype(x_dtype, np.unsignedinteger):
            if x_dtype == np.uint16:
                return AbstractRustDownsampler._switch_mod_with_y(
                    y_dtype, mod, f"{DOWNSAMPLE_F}_u16"
                )
            elif x_dtype == np.uint32:
                return AbstractRustDownsampler._switch_mod_with_y(
                    y_dtype, mod, f"{DOWNSAMPLE_F}_u32"
                )
            elif x_dtype == np.uint64:
                return AbstractRustDownsampler._switch_mod_with_y(
                    y_dtype, mod, f"{DOWNSAMPLE_F}_u64"
                )
        # INTS (need to be last because uint is subdtype of int)
        elif np.issubdtype(x_dtype, np.integer):
            if x_dtype == np.int16:
                return AbstractRustDownsampler._switch_mod_with_y(
                    y_dtype, mod, f"{DOWNSAMPLE_F}_i16"
                )
            elif x_dtype == np.int32:
                return AbstractRustDownsampler._switch_mod_with_y(
                    y_dtype, mod, f"{DOWNSAMPLE_F}_i32"
                )
            elif x_dtype == np.int64:
                return AbstractRustDownsampler._switch_mod_with_y(
                    y_dtype, mod, f"{DOWNSAMPLE_F}_i64"
                )
        # DATETIME -> i64 (datetime64 is viewed as int64)
        # TIMEDELTA -> i64 (timedelta64 is viewed as int64)
        raise ValueError(f"Unsupported data type (for x): {x_dtype}")

    def _downsample(
        self,
        x: Union[np.ndarray, None],
        y: np.ndarray,
        n_out: int,
        parallel: bool = False,
        **kwargs,
    ) -> np.ndarray:
        """Downsample the data in x and y."""
        mod = self.mod_single_core
        if parallel:
            if self.mod_multi_core is None:
                name = self.__class__.__name__
                warnings.warn(
                    f"No parallel implementation available for {name}. "
                    "Falling back to single-core implementation."
                )
            else:
                mod = self.mod_multi_core
        ## Viewing the y-data as different dtype (if necessary)
        if y.dtype == "bool":
            # bool is viewed as int8
            y = y.view(dtype=np.int8)
        elif np.issubdtype(y.dtype, np.datetime64):
            # datetime64 is viewed as int64
            y = y.view(dtype=np.int64)
        elif np.issubdtype(y.dtype, np.timedelta64):
            # timedelta64 is viewed as int64
            y = y.view(dtype=np.int64)
        ## Viewing the x-data as different dtype (if necessary)
        if x is None:
            downsample_f = self._switch_mod_with_y(y.dtype, mod)
            return downsample_f(y, n_out, **kwargs)
        elif np.issubdtype(x.dtype, np.datetime64):
            # datetime64 is viewed as int64
            x = x.view(dtype=np.int64)
        elif np.issubdtype(x.dtype, np.timedelta64):
            # timedelta64 is viewed as int64
            x = x.view(dtype=np.int64)
        ## Getting the appropriate downsample function
        downsample_f = self._switch_mod_with_x_and_y(x.dtype, y.dtype, mod)
        return downsample_f(x, y, n_out, **kwargs)

    def downsample(
        self,
        *args,  # x and y are optional
        n_out: int,
        parallel: bool = False,
        **kwargs,
    ):
        """Downsample the data in x and y."""
        return super().downsample(*args, n_out=n_out, parallel=parallel, **kwargs)



import numpy as np
import pytest
from test_config import supported_dtypes_x, supported_dtypes_y

from tsdownsample import (  # MeanDownsampler,; MedianDownsampler,
    EveryNthDownsampler,
    LTTBDownsampler,
    M4Downsampler,
    MinMaxDownsampler,
    MinMaxLTTBDownsampler,
)

# TODO: Improve tests
#   - compare implementations with existing plotly_resampler implementations


def test_m4_downsampler():
    """Test M4 downsampler."""
    arr = np.array(np.arange(10_000))
    s_downsampled = M4Downsampler().downsample(arr, n_out=100)
    assert s_downsampled[0] == 0
    assert s_downsampled[-1] == len(arr) - 1


def test_minmax_downsampler():
    """Test MinMax downsampler."""
    arr = np.array(np.arange(10_000))
    s_downsampled = MinMaxDownsampler().downsample(arr, n_out=100)
    assert s_downsampled[0] == 0
    assert s_downsampled[-1] == len(arr) - 1


def test_lttb_downsampler():
    """Test LTTB downsampler."""
    arr = np.array(np.arange(10_000))
    s_downsampled = LTTBDownsampler().downsample(arr, n_out=100)
    assert s_downsampled[0] == 0
    assert s_downsampled[-1] == len(arr) - 1


def test_minmaxlttb_downsampler():
    """Test MinMaxLTTB downsampler."""
    arr = np.array(np.arange(10_000))
    s_downsampled = MinMaxLTTBDownsampler().downsample(arr, n_out=100)
    assert s_downsampled[0] == 0
    assert s_downsampled[-1] == len(arr) - 1


def test_everynth_downsampler():
    """Test EveryNth downsampler."""
    arr = np.array(np.arange(10_000))
    s_downsampled = EveryNthDownsampler().downsample(arr, n_out=100)
    assert s_downsampled[0] == 0
    assert s_downsampled[-1] == 9_900


## Parallel downsampling

rust_downsamplers = [
    MinMaxDownsampler(),
    M4Downsampler(),
    LTTBDownsampler(),
    MinMaxLTTBDownsampler(),
]


def test_parallel_downsampling():
    """Test parallel downsampling."""
    arr = np.random.randn(10_000).astype(np.float32)
    for downsampler in rust_downsamplers:
        s_downsampled = downsampler.downsample(arr, n_out=100, parallel=False)
        s_downsampled_p = downsampler.downsample(arr, n_out=100, parallel=True)
        assert np.all(s_downsampled == s_downsampled_p)


def test_parallel_downsampling_with_x():
    """Test parallel downsampling with x."""
    arr = np.random.randn(10_001).astype(np.float32)  # 10_001 to test edge case
    idx = np.arange(len(arr))
    for downsampler in rust_downsamplers:
        s_downsampled = downsampler.downsample(idx, arr, n_out=100, parallel=False)
        s_downsampled_p = downsampler.downsample(idx, arr, n_out=100, parallel=True)
        assert np.all(s_downsampled == s_downsampled_p)


## Using x

all_downsamplers = rust_downsamplers + [EveryNthDownsampler()]


def test_downsampling_with_x():
    """Test downsampling with x."""
    arr = np.random.randn(2_001).astype(np.float32)  # 2_001 to test edge case
    idx = np.arange(len(arr))
    for downsampler in all_downsamplers:
        s_downsampled = downsampler.downsample(arr, n_out=100)
        s_downsampled_x = downsampler.downsample(idx, arr, n_out=100)
        assert np.all(s_downsampled == s_downsampled_x)


## Data types


def test_downsampling_different_dtypes():
    """Test downsampling with different data types."""
    arr_orig = np.random.randint(0, 100, size=10_000)
    for downsampler in rust_downsamplers:
        res = []
        for dtype in supported_dtypes_y:
            arr = arr_orig.astype(dtype)
            s_downsampled = downsampler.downsample(arr, n_out=100)
            if dtype is not np.bool_:
                res += [s_downsampled]
        for i in range(1, len(res)):
            assert np.all(res[0] == res[i])


def test_downsampling_different_dtypes_with_x():
    """Test downsampling with different data types."""
    arr_orig = np.random.randint(0, 100, size=10_000)
    idx_orig = np.arange(len(arr_orig))
    for downsampler in rust_downsamplers:
        for dtype_x in supported_dtypes_x:
            res = []
            idx = idx_orig.astype(dtype_x)
            for dtype_y in supported_dtypes_y:
                arr = arr_orig.astype(dtype_y)
                s_downsampled = downsampler.downsample(idx, arr, n_out=100)
                if dtype_y is not np.bool_:
                    res += [s_downsampled]
            for i in range(1, len(res)):
                assert np.all(res[0] == res[i])


### Check no out of bounds indexing


def test_downsampling_no_out_of_bounds_different_dtypes():
    """Test no out of bounds issues when downsampling with different data types."""
    arr_orig = np.random.randint(0, 100, size=100)
    for downsampler in rust_downsamplers:
        res = []
        for dtype in supported_dtypes_y:
            arr = arr_orig.astype(dtype)
            s_downsampled = downsampler.downsample(arr, n_out=76)
            s_downsampled_p = downsampler.downsample(arr, n_out=76, parallel=True)
            assert np.all(s_downsampled == s_downsampled_p)
            if dtype is not np.bool_:
                res += [s_downsampled]
        for i in range(1, len(res)):
            assert np.all(res[0] == res[i])


def test_downsampling_no_out_of_bounds_different_dtypes_with_x():
    """Test no out of bounds issues when downsampling with different data types."""
    arr_orig = np.random.randint(0, 100, size=100)
    idx_orig = np.arange(len(arr_orig))
    for downsampler in rust_downsamplers:
        for dtype_x in supported_dtypes_x:
            res = []
            idx = idx_orig.astype(dtype_x)
            for dtype_y in supported_dtypes_y:
                arr = arr_orig.astype(dtype_y)
                s_downsampled = downsampler.downsample(idx, arr, n_out=76)
                s_downsampled_p = downsampler.downsample(
                    idx, arr, n_out=76, parallel=True
                )
                assert np.all(s_downsampled == s_downsampled_p)
                if dtype_y is not np.bool_:
                    res += [s_downsampled]
            for i in range(1, len(res)):
                assert np.all(res[0] == res[i])


### Check no overflow when calculating average


def test_lttb_no_overflow():
    """Test no overflow when calculating average."""
    ### THIS SHOULD NOT OVERFLOW & HAVE THE SAME RESULT
    arr_orig = np.array([2 * 10**5] * 10_000, dtype=np.float64)
    s_downsampled = LTTBDownsampler().downsample(arr_orig, n_out=100)
    arr = arr_orig.astype(np.float32)
    s_downsampled_f32 = LTTBDownsampler().downsample(arr, n_out=100)
    assert np.all(s_downsampled == s_downsampled_f32)
    ### THIS SHOULD OVERFLOW & THUS HAVE A DIFFERENT RESULT...
    # max float32 is 3.4028235 × 1038 (so 2*10**38 is too big when adding 2 values)
    arr_orig = np.array([2 * 10**38] * 10_000, dtype=np.float64)
    s_downsampled = LTTBDownsampler().downsample(arr_orig, n_out=100)
    arr = arr_orig.astype(np.float32)
    s_downsampled_f32 = LTTBDownsampler().downsample(arr, n_out=100)
    assert not np.all(s_downsampled == s_downsampled_f32)  # TODO :(
    # I will leave this test here, but as many (much larger) libraries do not
    # really account for this, I guess it is perhaps less of an issue than I
    # thought. In the end f32 MAX is 3.4028235 × 1038 & f64 MAX is
    # 1.7976931348623157 × 10308 => which is in the end quite a lot.. (and all
    # integer averages are handled using f64) - f32 is only used for f16 & f32
    # (just as in numpy).


### Invalid n_out


def test_invalid_nout():
    """Test invalid n_out."""
    arr = np.random.randint(0, 100, size=10_000)
    with pytest.raises(ValueError):
        LTTBDownsampler().downsample(arr, n_out=-1)
    with pytest.raises(ValueError):
        # Should be even
        MinMaxDownsampler().downsample(arr, n_out=33)
    with pytest.raises(ValueError):
        # Should be multiple of 4
        M4Downsampler().downsample(arr, n_out=34)


### Unsupported dtype


def test_error_unsupported_dtype():
    """Test unsupported dtype."""
    arr = np.random.randint(0, 100, size=10_000)
    arr = arr.astype("object")
    with pytest.raises(ValueError):
        MinMaxDownsampler().downsample(arr, n_out=100)


def test_error_invalid_args():
    """Test invalid arguments."""
    arr = np.random.randint(0, 100, size=10_000)
    # No args
    with pytest.raises(ValueError) as e_msg:
        MinMaxDownsampler().downsample(n_out=100, parallel=True)
    assert "takes 1 or 2 positional arguments" in str(e_msg.value)
    # Too many args
    with pytest.raises(ValueError) as e_msg:
        MinMaxDownsampler().downsample(arr, arr, arr, n_out=100, parallel=True)
    assert "takes 1 or 2 positional arguments" in str(e_msg.value)
    # Invalid y
    with pytest.raises(ValueError) as e_msg:
        MinMaxDownsampler().downsample(arr.reshape(5, 2_000), n_out=100, parallel=True)
    assert "y must be 1D" in str(e_msg.value)
    # Invalid x
    with pytest.raises(ValueError) as e_msg:
        MinMaxDownsampler().downsample(
            arr.reshape(5, 2_000), arr, n_out=100, parallel=True
        )
    assert "x must be 1D" in str(e_msg.value)
    # Invalid x and y (different length)
    with pytest.raises(ValueError) as e_msg:
        MinMaxDownsampler().downsample(arr, arr[:-1], n_out=100, parallel=True)
    assert "x and y must have the same length" in str(e_msg.value)



"""tsdownsample: high performance downsampling of time series data for visualization."""

from .downsamplers import (
    EveryNthDownsampler,
    LTTBDownsampler,
    M4Downsampler,
    MinMaxDownsampler,
    MinMaxLTTBDownsampler,
)

__version__ = "0.1.0"
__author__ = "Jeroen Van Der Donckt"

__all__ = [
    "EveryNthDownsampler",
    "MinMaxDownsampler",
    "M4Downsampler",
    "LTTBDownsampler",
    "MinMaxLTTBDownsampler",
]
